import {
  BANK_INITIAL_REFRESH_TOKEN,
  TRANSACTION_RELATIVE_FROM_DATE,
  TRANSACTION_RELATIVE_TO_DATE,
} from "@/../config.ts"
import logger from "@/logger.ts"
import dayjs from "dayjs"
import { Database } from "better-sqlite3"
import {
  fetchToken,
  insertTokens,
  type TokenResponse,
} from "@/bank/db/queries.ts"
import * as Api from "./sparebank1Api.ts"

export interface OAuthTokenResponse {
  access_token: string
  expires_in: number
  refresh_token_expires_in: number
  refresh_token_absolute_expires_in: number
  token_type: "Bearer"
  refresh_token: string
}

export type BookingStatus = "PENDING" | "BOOKED"

export interface Transaction {
  id: string
  nonUniqueId: string
  date: number // Unix time
  amount: number // Amount in NOK
  cleanedDescription: string
  remoteAccountName: string
  bookingStatus: BookingStatus

  [key: string]: string | number | boolean | unknown
}

export interface TransactionResponse {
  transactions: ReadonlyArray<Transaction>
}

export interface Bank {
  transactionsPastDay: (
    ...accountKeys: ReadonlyArray<string>
  ) => Promise<TransactionResponse>
}

export class Sparebank1Impl implements Bank {
  private readonly db: Database

  constructor(db: Database) {
    this.db = db
  }

  private async getAccessToken(): Promise<string> {
    const accessToken = fetchToken(this.db, "access-token")

    if (accessToken && this.isValidToken(accessToken)) {
      return accessToken.token
    }
    const response = await this.fetchNewTokens()
    return response.access_token
  }

  private isValidToken(tokenResponse: TokenResponse): boolean {
    // TODO make sure the same timezone is used. Db uses UTC
    return dayjs().isBefore(tokenResponse.expires_at)
  }

  private async getRefreshToken(): Promise<string> {
    const tokenResponse = fetchToken(this.db, "refresh-token")

    if (!tokenResponse) {
      return BANK_INITIAL_REFRESH_TOKEN
    } else if (this.isValidToken(tokenResponse)) {
      return tokenResponse.token
    }
    // TODO clear database, if refresh token is invalid, will cause Exceptions on each call
    throw new Error("Refresh token is expired. Create a new one")
  }

  async fetchNewTokens(): Promise<OAuthTokenResponse> {
    const refreshToken = await this.getRefreshToken()
    const result = await Api.refreshToken(refreshToken)

    if (result.status === "failure") {
      throw logger.error({
        err: new Error(`Failed to fetch refresh token: '${result.data}'`),
      })
    }
    const oAuthToken = result.data

    insertTokens(this.db, oAuthToken)
    return oAuthToken
  }

  async transactionsPastDay(
    ...accountKeys: ReadonlyArray<string>
  ): Promise<TransactionResponse> {
    const today = dayjs()
    const fromDate = today.subtract(TRANSACTION_RELATIVE_FROM_DATE, "days")
    const toDate = today.subtract(TRANSACTION_RELATIVE_TO_DATE, "days")
    return await Api.transactions(await this.getAccessToken(), accountKeys, {
      fromDate,
      toDate,
    })
  }
}
